"""Adopt environment section in pytest configuration files."""

from __future__ import annotations

import os
import sys
from dataclasses import dataclass
from itertools import chain
from typing import Iterator

import pytest

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    import tomllib
else:  # pragma: <3.11 cover
    import tomli as tomllib


def pytest_addoption(parser: pytest.Parser) -> None:
    """Add section to configuration files."""
    help_msg = "a line separated list of environment variables of the form (FLAG:)NAME=VALUE"
    parser.addini("env", type="linelist", help=help_msg, default=[])


@dataclass
class Entry:
    """Configuration entries."""

    key: str
    value: str
    transform: bool
    skip_if_set: bool


@pytest.hookimpl(tryfirst=True)
def pytest_load_initial_conftests(
    args: list[str],  # noqa: ARG001
    early_config: pytest.Config,
    parser: pytest.Parser,  # noqa: ARG001
) -> None:
    """Load environment variables from configuration files."""
    for entry in _load_values(early_config):
        if entry.skip_if_set and entry.key in os.environ:
            continue
        # transformation -> replace environment variables, e.g. TEST_DIR={USER}/repo_test_dir.
        os.environ[entry.key] = entry.value.format(**os.environ) if entry.transform else entry.value


def _load_values(early_config: pytest.Config) -> Iterator[Entry]:
    has_toml_conf = False
    for path in chain.from_iterable([[early_config.rootpath], early_config.rootpath.parents]):  # noqa: PLR1702
        toml_file = path / "pyproject.toml"
        if toml_file.exists():
            with toml_file.open("rb") as file_handler:
                config = tomllib.load(file_handler)
                if "tool" in config and "pytest_env" in config["tool"]:
                    has_toml_conf = True
                    for key, entry in config["tool"]["pytest_env"].items():
                        if isinstance(entry, dict):
                            value = str(entry["value"])
                            transform, skip_if_set = bool(entry.get("transform")), bool(entry.get("skip_if_set"))
                        else:
                            value, transform, skip_if_set = str(entry), False, False
                        yield Entry(key, value, transform, skip_if_set)
            break

    if not has_toml_conf:
        for line in early_config.getini("env"):
            # INI lines e.g. D:R:NAME=VAL has two flags (R and D), NAME key, and VAL value
            parts = line.partition("=")
            ini_key_parts = parts[0].split(":")
            flags = {k.strip().upper() for k in ini_key_parts[:-1]}
            # R: is a way to designate whether to use raw value -> perform no transformation of the value
            transform = "R" not in flags
            # D: is a way to mark the value to be set only if it does not exist yet
            skip_if_set = "D" in flags
            key = ini_key_parts[-1].strip()
            value = parts[2].strip()
            yield Entry(key, value, transform, skip_if_set)



from __future__ import annotations

import ast
import os


def test_env() -> None:
    for key, value in ast.literal_eval(os.environ["_TEST_ENV"]).items():
        assert os.environ[key] == value, key



from __future__ import annotations

import os
import re
from pathlib import Path
from unittest import mock

import pytest


@pytest.mark.parametrize(
    ("env", "ini", "expected_env"),
    [
        pytest.param(
            {},
            "[pytest]\nenv = MAGIC=alpha",
            {"MAGIC": "alpha"},
            id="new key - add to env",
        ),
        pytest.param(
            {},
            "[pytest]\nenv = MAGIC=alpha\n SORCERY=beta",
            {"MAGIC": "alpha", "SORCERY": "beta"},
            id="two new keys - add to env",
        ),
        pytest.param(
            # This test also tests for non-interference of env variables between this test and tests above
            {},
            "[pytest]\nenv = d:MAGIC=beta",
            {"MAGIC": "beta"},
            id="D flag - add to env",
        ),
        pytest.param(
            {"MAGIC": "alpha"},
            "[pytest]\nenv = MAGIC=beta",
            {"MAGIC": "beta"},
            id="key exists in env - overwrite",
        ),
        pytest.param(
            {"MAGIC": "alpha"},
            "[pytest]\nenv = D:MAGIC=beta",
            {"MAGIC": "alpha"},
            id="D exists - original val kept",
        ),
        pytest.param(
            {"PLANET": "world"},
            "[pytest]\nenv = MAGIC=hello_{PLANET}",
            {"MAGIC": "hello_world"},
            id="curly exist - interpolate var",
        ),
        pytest.param(
            {"PLANET": "world"},
            "[pytest]\nenv = R:MAGIC=hello_{PLANET}",
            {"MAGIC": "hello_{PLANET}"},
            id="R exists - not interpolate var",
        ),
        pytest.param(
            {"MAGIC": "a"},
            "[pytest]\nenv = R:MAGIC={MAGIC}b\n D:MAGIC={MAGIC}c\n MAGIC={MAGIC}d",
            {"MAGIC": "{MAGIC}bd"},
            id="incremental interpolation",
        ),
        pytest.param(
            {"PLANET": "world"},
            "[pytest]\nenv = D:R:RESULT=hello_{PLANET}",
            {"RESULT": "hello_{PLANET}"},
            id="two flags",
        ),
        pytest.param(
            {"PLANET": "world"},
            "[pytest]\nenv = R:D:RESULT=hello_{PLANET}",
            {"RESULT": "hello_{PLANET}"},
            id="two flags - reversed",
        ),
        pytest.param(
            {"PLANET": "world"},
            "[pytest]\nenv = d:r:RESULT=hello_{PLANET}",
            {"RESULT": "hello_{PLANET}"},
            id="lowercase flags",
        ),
        pytest.param(
            {"PLANET": "world"},
            "[pytest]\nenv =  D  :  R  :  RESULT  =  hello_{PLANET}",
            {"RESULT": "hello_{PLANET}"},
            id="whitespace is ignored",
        ),
        pytest.param(
            {"MAGIC": "zero"},
            "",
            {"MAGIC": "zero"},
            id="empty ini works",
        ),
    ],
)
def test_env_via_pytest(
    testdir: pytest.Testdir,
    env: dict[str, str],
    ini: str,
    expected_env: dict[str, str],
    request: pytest.FixtureRequest,
) -> None:
    tmp_dir = Path(str(testdir.tmpdir))
    test_name = re.sub(r"\W|^(?=\d)", "_", request.node.callspec.id).lower()
    Path(str(tmp_dir / f"test_{test_name}.py")).symlink_to(Path(__file__).parent / "template.py")
    (tmp_dir / "pytest.ini").write_text(ini, encoding="utf-8")

    new_env = {
        **env,
        "_TEST_ENV": repr(expected_env),
        "PYTEST_DISABLE_PLUGIN_AUTOLOAD": "1",
        "PYTEST_PLUGINS": "pytest_env.plugin",
    }

    # monkeypatch persists env variables across parametrized tests, therefore using mock.patch.dict
    with mock.patch.dict(os.environ, new_env, clear=True):
        result = testdir.runpytest()

    result.assert_outcomes(passed=1)


@pytest.mark.parametrize(
    ("env", "toml", "ini", "expected_env"),
    [
        pytest.param(
            {},
            '[tool.pytest.ini_options]\nenv = ["MAGIC=toml", "MAGIC_2=toml2"]',
            "[pytest]\nenv = MAGIC=ini\n MAGIC_2=ini2",
            {"MAGIC": "ini", "MAGIC_2": "ini2"},
            id="ini over toml ini_options",
        ),
        pytest.param(
            {},
            '[tool.pytest.ini_options]\nenv = ["MAGIC=toml", "MAGIC_2=toml2"]',
            "",
            {"MAGIC": "toml", "MAGIC_2": "toml2"},
            id="toml via ini_options",
        ),
        pytest.param(
            {},
            '[tool.pytest_env]\nMAGIC = 1\nMAGIC_2 = "toml2"',
            "",
            {"MAGIC": "1", "MAGIC_2": "toml2"},
            id="toml native",
        ),
        pytest.param(
            {},
            '[tool.pytest_env]\nMAGIC = 1\nMAGIC_2 = "toml2"',
            "[pytest]\nenv = MAGIC=ini\n MAGIC_2=ini2",
            {"MAGIC": "1", "MAGIC_2": "toml2"},
            id="toml native over ini",
        ),
        pytest.param(
            {},
            '[tool.pytest_env]\nMAGIC = {value = "toml", "transform"= true, "skip_if_set" = true}',
            "",
            {"MAGIC": "toml"},
            id="toml inline table",
        ),
    ],
)
def test_env_via_toml(  # noqa: PLR0913, PLR0917
    testdir: pytest.Testdir,
    env: dict[str, str],
    toml: str,
    ini: str,
    expected_env: dict[str, str],
    request: pytest.FixtureRequest,
) -> None:
    tmp_dir = Path(str(testdir.tmpdir))
    test_name = re.sub(r"\W|^(?=\d)", "_", request.node.callspec.id).lower()
    Path(str(tmp_dir / f"test_{test_name}.py")).symlink_to(Path(__file__).parent / "template.py")
    if ini:
        (tmp_dir / "pytest.ini").write_text(ini, encoding="utf-8")
    (tmp_dir / "pyproject.toml").write_text(toml, encoding="utf-8")

    new_env = {
        **env,
        "_TEST_ENV": repr(expected_env),
        "PYTEST_DISABLE_PLUGIN_AUTOLOAD": "1",
        "PYTEST_PLUGINS": "pytest_env.plugin",
    }

    # monkeypatch persists env variables across parametrized tests, therefore using mock.patch.dict
    with mock.patch.dict(os.environ, new_env, clear=True):
        result = testdir.runpytest()

    result.assert_outcomes(passed=1)


def test_env_via_toml_bad(testdir: pytest.Testdir) -> None:
    toml_file = Path(str(testdir.tmpdir)) / "pyproject.toml"
    toml_file.write_text("bad toml", encoding="utf-8")

    result = testdir.runpytest()
    assert result.ret == 4
    assert result.errlines == [
        f"ERROR: {toml_file}: Expected '=' after a key in a key/value pair (at line 1, column 5)",
        "",
    ]
